#include <stdint.h>
#include <libxml/parser.h>
#include <string.h>
#include <inttypes.h>
#include "xml.h"
#include "logger.h"

void init_xml(void)
{
  xmlInitMemory();
  LIBXML_TEST_VERSION
}

static char *get_property(xmlAttrPtr attr, char *p)
{
  if (attr == NULL) return "";
  if (!strcmp((char*)attr->name, p)) return (char*)attr->children->content;
  return get_property(attr->next, p);
}

static uint64_t get_start(xmlNodePtr doc)
{
  char *p;
  if (doc == NULL) return 0;
  if (strcmp((char*)doc->name, "c") != 0) return get_start(doc->next);
  p = get_property(doc->properties, "t");
  if (p[0]) {
    uint64_t ret;
    if (sscanf(p, "%"SCNu64, &ret) != 1) return 0;
    return ret;
  }
  return get_start(doc->next);
}

static uint64_t get_inc(xmlNodePtr doc)
{
  char *p;
  if (doc == NULL) return 0;
  if (strcmp((char*)doc->name, "c") != 0) return get_inc(doc->next);
  p = get_property(doc->properties, "d");
  if (p[0]) {
    uint64_t ret;
    if (sscanf(p, "%"SCNu64, &ret) != 1) return 0;
    return ret;
  }
  return get_inc(doc->next);
}

static void get_best_quality(xmldata *d, xmlNodePtr doc, int max_width)
{
  if (doc == NULL) return;
  if (!strcmp((char*)doc->name, "QualityLevel")) {
    uint64_t bitrate;
    int width;
    sscanf(get_property(doc->properties, "Bitrate"), "%"SCNu64, &bitrate);
    sscanf(get_property(doc->properties, "MaxWidth"), "%d", &width);
    printf("INFO: stream width %d\n", width);
    if ((max_width > 0 && width <= max_width && width > d->width) ||
        (max_width <= 0 && bitrate > d->video_bitrate) ||
        (d->width == 0)) {
      d->width = width;
      d->video_bitrate = bitrate;
      d->video_priv = get_property(doc->properties, "CodecPrivateData");
    }
  }
  get_best_quality(d, doc->next, max_width);
}

static void get_audio_quality(xmldata *d, xmlNodePtr doc)
{
  if (doc == NULL) return;
  if (!strcmp((char*)doc->name, "QualityLevel")) {
    uint64_t bitrate;
    sscanf(get_property(doc->properties, "Bitrate"), "%"SCNu64, &bitrate);
    if (bitrate > d->audio_bitrate) {
      d->audio_bitrate = bitrate;
      d->audio_priv = get_property(doc->properties, "CodecPrivateData");
    }
  }
  get_audio_quality(d, doc->next);
}

struct timings {
  uint64_t *t;
  int tsize;
};

void add_time(struct timings *r, uint64_t t)
{
  if ((r->tsize & 127) == 0) {
    r->t = realloc(r->t, (r->tsize + 128) * sizeof(uint64_t));
    if (r->t == NULL) abort();
  }
  r->t[r->tsize] = t;
  r->tsize++;
}

static struct timings get_timings(xmlNodePtr doc)
{
  struct timings ret = { NULL, 0 };
  uint64_t t = 0;
  uint64_t d = 0;

  while (doc != NULL) {
    xmlAttrPtr attr;
    if (strcmp((char*)doc->name, "c")) { doc = doc->next; continue; }
    attr = doc->properties;
    while (attr) {
      char *p = (char*)attr->name;
      char *v = (char*)attr->children->content;
      if (!strcmp(p, "d")) if (sscanf(v, "%"SCNu64, &d) != 1) abort();
      if (!strcmp(p, "t")) if (sscanf(v, "%"SCNu64, &t) != 1) abort();
      attr = attr->next;
    }
    add_time(&ret, t);
    t += d;
    doc = doc->next;
  }
  add_time(&ret, t);

  return ret;
}

static void get_video(xmldata *d, xmlNodePtr doc, int max_width)
{
  uint64_t timescale;
  struct timings t;
  if (doc == NULL) return;
  /* get default TimeScale, if exists, video will overwrite if exists */
  if (!strcmp((char*)doc->name, "SmoothStreamingMedia")) {
    if (sscanf(get_property(doc->properties, "TimeScale"), "%"SCNu64,
               &timescale) == 1) {
      if (timescale != 10000000)
        printf("WARNING: TimeScale not 10000000 (%"PRIu64")\n", timescale);
      d->video_timescale = timescale;
    }
  }
  if (!strcmp((char*)doc->name, "StreamIndex") &&
      !strcmp(get_property(doc->properties, "Type"), "video")) {
    d->video_url = get_property(doc->properties, "Url");
    if (sscanf(get_property(doc->properties, "TimeScale"), "%"SCNu64,
               &timescale) == 1) {
      if (timescale != 10000000)
        printf("WARNING: video TimeScale not 10000000 (%"PRIu64")\n",
               timescale);
      d->video_timescale = timescale;
    }
    t = get_timings(doc->children);
    d->video_times = t.t;
    d->video_times_size = t.tsize;
    d->video_start = get_start(doc->children);
    d->video_inc = get_inc(doc->children);
    if (d->video_inc != 20000000) {
      printf("WARNING: force video_inc to 20000000 "
             "(downloaded value %"PRIu64")\n", d->video_inc);
      d->video_inc = 20000000;
    }
    d->width = 0;
    d->video_bitrate = 0;
    d->video_priv = "";
    get_best_quality(d, doc->children, max_width);
    printf("INFO: chosen width %d (max allowed %d)\n", d->width, max_width);
    return;
  }
  get_video(d, doc->children, max_width);
  get_video(d, doc->next, max_width);
}

static void get_audio(xmldata *d, xmlNodePtr doc)
{
  uint64_t timescale;
  struct timings t;
  if (doc == NULL) return;
  /* get default TimeScale, if exists, audio will overwrite if exists */
  if (!strcmp((char*)doc->name, "SmoothStreamingMedia")) {
    if (sscanf(get_property(doc->properties, "TimeScale"), "%"SCNu64,
               &timescale) == 1) {
      if (timescale != 10000000)
        printf("WARNING: TimeScale not 10000000 (%"PRIu64")\n", timescale);
      d->audio_timescale = timescale;
    }
  }
  if (!strcmp((char*)doc->name, "StreamIndex") &&
      !strcmp(get_property(doc->properties, "Type"), "audio")) {
    d->audio_url = get_property(doc->properties, "Url");
    if (sscanf(get_property(doc->properties, "TimeScale"), "%"SCNu64,
               &timescale) == 1) {
      if (timescale != 10000000)
        printf("WARNING: audio TimeScale not 10000000 (%"PRIu64")\n",
               timescale);
      d->audio_timescale = timescale;
    }
    t = get_timings(doc->children);
    d->audio_times = t.t;
    d->audio_times_size = t.tsize;
    d->audio_start = get_start(doc->children);
    d->audio_inc = get_inc(doc->children);
    if (d->audio_inc != 19840000) {
      printf("WARNING: force audio_inc to 19840000 "
             "(downloaded value %"PRIu64")\n", d->audio_inc);
      d->audio_inc = 19840000;
    }
    d->audio_bitrate = 0;
    d->audio_priv = "";
    get_audio_quality(d, doc->children);
    return;
  }
  get_audio(d, doc->children);
  get_audio(d, doc->next);
}

void analyze_streams(xmldata *d, xmlNodePtr doc)
{
  if (doc == NULL) return;
  if (!strcmp((char*)doc->name, "SmoothStreamingMedia")) {
    d->smooth_streaming_media_count++;
    if (strcmp(get_property(doc->properties, "IsLive"), "TRUE"))
      printf("WARNING: stream is not 'IsLive'\n");
    if (strcmp(get_property(doc->properties, "LookAheadFragmentCount"), "2"))
      printf("WARNING: stream 'LookAheadFragmentCount' is not 2 (%s)\n",
             get_property(doc->properties, "LookAheadFragmentCount"));
    if (strcmp(get_property(doc->properties, "MajorVersion"), "2"))
      printf("WARNING: stream 'MajorVersion' is not 2 (%s)\n",
             get_property(doc->properties, "MajorVersion"));
    if (strcmp(get_property(doc->properties, "MinorVersion"), "0"))
      printf("WARNING: stream 'MinorVersion' is not 0 (%s)\n",
             get_property(doc->properties, "MinorVersion"));
  }
  if (!strcmp((char*)doc->name, "StreamIndex")) {
    if (!strcmp(get_property(doc->properties, "Type"), "video"))
      d->stream_index_video_count++;
    else if (!strcmp(get_property(doc->properties, "Type"), "audio"))
      d->stream_index_audio_count++;
    else printf("WARNING: unknown stream type '%s'\n",
                get_property(doc->properties, "Type"));
  }
  analyze_streams(d, doc->children);
  analyze_streams(d, doc->next);
}

int parse_xml(xmldata *d, char *data, int size, int max_width)
{
  xmlDocPtr doc = xmlParseMemory(data, size);

  if (doc == NULL) { logprintf("no XML data?\n"); return -1; }

  /* set default time scale to 10000000 (so says the spec) */
  d->video_timescale = 10000000;
  d->audio_timescale = 10000000;

  analyze_streams(d, doc->children);
  if (d->smooth_streaming_media_count != 1)
    printf("WARNING: 'SmoothStreamingMedia' count is"
           " expected to be 1 but is %d\n",
           d->smooth_streaming_media_count);
  if (d->smooth_streaming_media_count == 1) {
    if (d->stream_index_video_count != 1)
      printf("WARNING: video stream count is expected to be 1 but is %d\n",
           d->stream_index_video_count);
    if (d->stream_index_audio_count != 1)
      printf("WARNING: audio stream count is expected to be 1 but is %d\n",
           d->stream_index_audio_count);
  }

  get_video(d, doc->children, max_width);
  get_audio(d, doc->children);

  if (d->video_url != NULL) d->video_url = strdup(d->video_url);
  if (d->video_priv != NULL) d->video_priv = strdup(d->video_priv);
  //if (d->video_url == NULL || d->video_priv == NULL) abort();

  if (d->audio_url != NULL) d->audio_url = strdup(d->audio_url);
  if (d->audio_priv != NULL) d->audio_priv = strdup(d->audio_priv);
  //if (d->audio_url == NULL || d->audio_priv == NULL) abort();

  logprintf("video url %s\n", d->video_url);
  logprintf("video start %"PRIu64"\n", d->video_start);
  logprintf("video inc %"PRIu64"\n", d->video_inc);
  logprintf("video bitrate %"PRIu64"\n", d->video_bitrate);
  logprintf("video private data %s\n", d->video_priv);

  logprintf("audio url %s\n", d->audio_url);
  logprintf("audio start %"PRIu64"\n", d->audio_start);
  logprintf("audio inc %"PRIu64"\n", d->audio_inc);
  logprintf("audio bitrate %"PRIu64"\n", d->audio_bitrate);
  logprintf("audio private data %s\n", d->audio_priv);
  xmlFreeDoc(doc);

  return 0;
}

int try_parse_xml(char *data, int size)
{
  xmlDocPtr doc = xmlParseMemory(data, size);
  if (doc == NULL) { logprintf("no XML data?\n"); return -1; }
  xmlFreeDoc(doc);
  return 0;
}

#if 0
int main(void)
{
  init_xml();
  xmldata d; memset(&d, 0, sizeof(d));
  char data[4096];
  FILE *f = fopen("Manifest?IS=0&ET=1482681401&CIP=90.8.126.206&KO=1&KN=4&US=3e39b036652bf140a7c24ea77147d2c0&SID=OTV-243705bd-29cc-49dd-a940-968af6c11293", "r");
  int size = fread(data, 1, 4096, f); fclose(f);
  int ret = parse_xml(&d, data, size, -1);
if (1) data[0] = 1;
  return 0;
}
#endif
